/*
 * VIDNS - Virtual Internal DNS management system
 * Copyright (C) 2022 lemon.sh
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of  MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

use crate::config::Config;
use crate::storage::StorageService;
use crate::web::WebBackend;
use std::str::FromStr;
use std::sync::Arc;
use std::{env, fs, thread};
use tokio::signal::unix::{signal, SignalKind};
use tokio::sync::broadcast;
use tokio::sync::mpsc::unbounded_channel;
use tracing::Level;

mod config;
mod storage;
mod web;

#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
    let config_var = env::var("VIDNS_CONFIG");
    let config_path = config_var.as_deref().unwrap_or("vidns.toml");
    println!("Loading config from '{}'...", config_path);
    let config_str = fs::read_to_string(config_path)?;
    let config: Arc<Config> = Arc::new(toml::from_str(&config_str)?);

    tracing_subscriber::fmt::fmt()
        .with_max_level({
            if let Some(o) = config.log_level.as_deref() {
                Level::from_str(o)?
            } else {
                Level::INFO
            }
        })
        .init();

    tracing::info!(concat!("Initializing - VIDNS v", env!("CARGO_PKG_VERSION")));
    let (ctx, _) = broadcast::channel(1);
    let (etx, mut erx) = unbounded_channel();

    tracing::info!(
        "Started web backend, listening at http://{}",
        config.web.listen
    );

    let nsdcpath = config
        .nsd
        .nsd_control_binary
        .as_deref()
        .unwrap_or("/usr/bin/nsd-control");
    let (storage_service, storage) = StorageService::create(&config.database, nsdcpath)?;
    let storage_thread = thread::spawn(|| storage_service.run());

    let web_backend = WebBackend::new(config, storage).run()?;
    let web_task = tokio::spawn(async move {
        if let Err(e) = web_backend.await {
            let _err = etx.send(e);
        }
    });

    let mut sigterm = signal(SignalKind::terminate()).unwrap();
    let mut sigint = signal(SignalKind::interrupt()).unwrap();
    tokio::select! {
        _ = sigterm.recv() => tracing::info!("SIGTERM signal received, powering down"),
        _ = sigint.recv() => tracing::info!("SIGINT signal received, powering down"),
        e = erx.recv() => {
            if let Some(e) = e {
                tracing::error!("An error has occurred, shutting down: {}", e);
            } else {
                tracing::error!("Error channel has been dropped due to an unknown error, shutting down");
            }
        }
    }

    let _ = ctx.send(());
    web_task.await.unwrap();
    storage_thread.join().unwrap();

    tracing::info!("Shutdown complete");
    Ok(())
}
