/*
 * VIDNS - Virtual Internal DNS management system
 * Copyright (C) 2022 lemon.sh
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of  MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

use std::{path::PathBuf, str::FromStr};

use anyhow::{anyhow, Result};
use rusqlite::{ErrorCode, OptionalExtension};
use tokio::sync::{
    mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender},
    oneshot,
};
use tokio::time::Instant;

mod nsdcontrol;

macro_rules! executor_wrapper {
    ($name:ident, $task:expr, $ret:ty, $($arg:ident: $ty:ty),*) => {
        pub async fn $name(&self, $($arg: $ty),*) -> $ret {
          let (otx, orx) = oneshot::channel();
          self.tx.send($task(otx, $($arg),*)).unwrap();
          orx.await.unwrap()
        }
    };
    ($name:ident, $task:expr, $ret:ty) => {
        pub async fn $name(&self) -> $ret {
          let (otx, orx) = oneshot::channel();
          self.tx.send($task(otx)).unwrap();
          orx.await.unwrap()
        }
    };
}

pub struct StorageService {
    rx: UnboundedReceiver<Task>,
    db: rusqlite::Connection,
    nsdcpath: PathBuf,
}

impl StorageService {
    pub fn create(dbpath: &str, nsdcpath: &str) -> anyhow::Result<(Self, StorageConnection)> {
        let (tx, rx) = unbounded_channel();
        let db = rusqlite::Connection::open(dbpath)?;
        db.execute("pragma foreign_keys = on", [])?;
        db.execute_batch(include_str!("schema.sql"))?;
        tracing::debug!("Database loaded ({})", dbpath);
        Ok((
            Self {
                rx,
                db,
                nsdcpath: PathBuf::from_str(nsdcpath)?,
            },
            StorageConnection { tx },
        ))
    }

    pub fn run(mut self) {
        while let Some(task) = self.rx.blocking_recv() {
            let before = Instant::now();
            tracing::debug!("got task {:?}", task);
            match task {
                Task::RegisterAccount(tx, u, h) => tx.send(self.register_account(u, h)).unwrap(),
                Task::GetAccountByUsername(tx, u) => {
                    tx.send(self.get_account_by_username(u)).unwrap()
                }
            }
            tracing::debug!(
                "task took {}ms",
                Instant::now().duration_since(before).as_secs_f64() / 1000.0
            );
        }
    }

    fn register_account(&self, username: String, hash: String) -> Result<Option<i64>> {
        let result = self.db.query_row(
            "insert into accounts(username,hash) values (?, ?) returning id",
            [username, hash],
            |v| v.get(0),
        );
        match result {
            Ok(o) => Ok(Some(o)),
            Err(err) => {
                if let rusqlite::Error::SqliteFailure(serr, _) = err && serr.code == ErrorCode::ConstraintViolation {
                    Ok(None)
                } else {
                    Err(anyhow!(err))
                }
            }
        }
    }

    fn get_account_by_username(&self, username: String) -> Result<Option<AccountCredentials>> {
        Ok(self
            .db
            .query_row(
                "select id,hash from accounts where username=?",
                [username],
                |v| {
                    Ok(AccountCredentials {
                        id: v.get::<_, i64>(0)?,
                        hash: v.get::<_, String>(1)?,
                    })
                },
            )
            .optional()?)
    }
}

pub struct StorageConnection {
    tx: UnboundedSender<Task>,
}

impl Clone for StorageConnection {
    fn clone(&self) -> Self {
        Self {
            tx: self.tx.clone(),
        }
    }
}

#[derive(Debug)]
pub struct AccountCredentials {
    pub id: i64,
    pub hash: String,
}

type ResultSender<T> = oneshot::Sender<Result<T>>;

#[derive(Debug)]
enum Task {
    RegisterAccount(ResultSender<Option<i64>>, String, String),
    GetAccountByUsername(ResultSender<Option<AccountCredentials>>, String),
}

impl StorageConnection {
    executor_wrapper!(
        register_account,
        Task::RegisterAccount,
        Result<Option<i64>>,
        username: String,
        hash: String
    );
    executor_wrapper!(
        get_account,
        Task::GetAccountByUsername,
        Result<Option<AccountCredentials>>,
        username: String
    );
}
