/*
 * VIDNS - Virtual Internal DNS management system
 * Copyright (C) 2022 lemon.sh
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of  MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

use crate::storage::StorageConnection;
use crate::Config;
use actix_session::storage::CookieSessionStore;
use actix_session::SessionMiddleware;
use actix_web::cookie::Key;
use actix_web::dev::Server;
use actix_web::web::Data;
use actix_web::{App, HttpServer};
use argon2::Argon2;
use std::sync::Arc;

mod forms;
mod routes;
mod templates;

pub struct WebBackend {
    config: Arc<Config>,
    db: StorageConnection,
}

impl WebBackend {
    pub fn new(config: Arc<Config>, db: StorageConnection) -> Self {
        Self { config, db }
    }

    pub fn run(self) -> anyhow::Result<Server> {
        let key_bytes = hex::decode(&self.config.web.cookie_secret)?;
        let key = Key::try_from(key_bytes.as_slice())?;
        let server = HttpServer::new(move || {
            App::new()
                .wrap(SessionMiddleware::new(
                    CookieSessionStore::default(),
                    key.clone(),
                ))
                .app_data(Data::new(Argon2::default()))
                .app_data(Data::new(self.db.clone()))
                .service(routes::login)
                .service(routes::login_post)
                .service(routes::register)
                .service(routes::register_post)
                .service(routes::static_files)
        })
        .bind(self.config.web.listen)?
        .run();
        Ok(server)
    }
}
