use actix_session::Session;
use actix_web::{web, HttpRequest, HttpResponse, Responder};
use argon2::{Argon2, PasswordHash, PasswordVerifier};
use rust_embed::RustEmbed;

use crate::{storage::StorageConnection, web::forms, web::templates};

#[derive(RustEmbed)]
#[folder = "static/"]
struct StaticFiles;

#[actix_web::get("/static/{_:.*}")]
pub async fn static_files(path: web::Path<String>) -> impl Responder {
    StaticFiles::get(&path).map(|f| {
        HttpResponse::Ok()
            .content_type(
                mime_guess::from_path(path.as_str())
                    .first_or_octet_stream()
                    .as_ref(),
            )
            .body(f.data.into_owned())
    })
}

#[actix_web::get("/")]
pub async fn login(session: Session, req: HttpRequest) -> actix_web::Result<impl Responder> {
    if let Some(_) = session.get::<i64>("uid")? {
        Ok(HttpResponse::TemporaryRedirect().finish())
    } else {
        Ok(templates::LoginTemplate { flash: None }.respond_to(&req))
    }
}

#[actix_web::get("/register")]
pub async fn register(
    session: actix_session::Session,
    req: HttpRequest,
) -> actix_web::Result<impl Responder> {
    if let Some(_) = session.get::<i64>("uid")? {
        Ok(HttpResponse::TemporaryRedirect().finish())
    } else {
        Ok(templates::RegisterTemplate { flash: None }.respond_to(&req))
    }
}

#[actix_web::post("/")]
pub async fn login_post(
    session: Session,
    db: web::Data<StorageConnection>,
    argon: web::Data<Argon2<'_>>,
    form: web::Form<forms::Login>,
    req: HttpRequest,
) -> actix_web::Result<impl Responder> {
    match db.get_account(form.login.clone()).await {
        Ok(o) => {
            if let Some(creds) = o {
                let hash = PasswordHash::new(&creds.hash).unwrap();
                if argon
                    .verify_password(&form.password.as_bytes(), &hash)
                    .is_err()
                {
                    return Ok(templates::LoginTemplate {
                        flash: Some("Username or password is invalid."),
                    }
                    .respond_to(&req));
                }
                session.insert("uid", creds.id)?;
                Ok(HttpResponse::Found().finish())
            } else {
                Ok(templates::LoginTemplate {
                    flash: Some("This account does not exist."),
                }
                .respond_to(&req))
            }
        }
        Err(e) => {
            tracing::error!("{}", e);
            Ok(templates::LoginTemplate {
                flash: Some("An unknown error has occurred."),
            }
            .respond_to(&req))
        }
    }
}

#[actix_web::post("/register")]
pub async fn register_post(
    session: Session,
    form: web::Form<forms::Login>,
) -> actix_web::Result<impl Responder> {
    Ok("")
}
